unit RASControl;

{
TRASControl component for Delphi (version 1.2)
----------------------------------------------
Copyright (c) 1999 by David A. Bussey


Version Info
------------

v1.0
  - Initial release.

v1.1
  - TRASControl.PhoneNumber/DeviceName/DeviceType now display values properly.
  - Improved RAS Callback function.

v1.2
  - Calls to DialEntry no longer remove phonebook entry's saved password.


There are still a few things that can be improved.  If you have any suggestions
for improvements please let me know your ideas.  If they are possible, I will
try and include them in later versions.

(you can reach me at dbussey@bluegrass.net)

TRASControl is FREEWARE.  It is supplied as is, without warranty of any kind.
Use it at your own risk.  Any components derived from TRASControl must also be
FREEWARE.  You may use TRASControl in both private and commercial applications
without any fees or royalties.  No person/persons may redistribute TRASControl
as their own or for any profit whatsoever.  If you redistribute TRASControl you
must supply the files in their original un-altered states.

This component was created using the RASAPI header file (version 2.11) ported
to Delphi by Davide Moretti (dave@rimini.com).  In order to keep cross-platform
compatability I have only included in this version of the component functions
which are common to 95(OSR2)/98/NT.  As far as pre-OSR2 versions of Windows 95,
the RNAPH dynamic link library must be supplied with the component and
application.  This DLL is available from Microsoft.

TRASControl was developed on a Windows NT 4.0 (SP-4) platform with
Delphi 4 C/S (SP-2).  I have not tried to use this component with earlier
versions of Delphi but I cannot imagine any problems with using it in
Delphi 2 or 3 off-hand.
}

interface

uses
  Windows, SysUtils, Classes, Ras;

const
  RASControl_Unknown = -1;

type
  TRASNotifyEvent = procedure(Sender: TObject; EntryIndex: integer; rasStatus: integer; rasError: Longint) of object;
  TRASCallBackProc = procedure(hConn: THRasConn; msg: Integer; state: TRasConnState; dwError: Longint; dwEexterror: Longint); stdcall;
  PRASCallBackProc = ^TRASCallBackProc;
  ERASControlError = class(Exception);

type
  TRASMonitorThread = class(TThread)
  private
    { Private declarations }
    FOwner: TComponent;
    FEntryIndex: integer;
    FRASState: TRasConnState;
    FRASError: Longint;
  protected
    { Protected declarations }
    procedure Execute; override;
    procedure RASStatusChange;
  public
    { Public declarations }
    RASCallBackProc: PRASCallBackProc;
    constructor Create(AOwner: TComponent; FSuspended: Boolean; Index: integer);
    destructor Destroy; override;
    procedure UpdateStatus;
    property EntryIndex: integer read FEntryIndex write FEntryIndex;
  end;

type
  PRASEntryItem = ^TRASEntryItem;
  TRASEntryItem = record
    EntryName: string;
    PhoneNumber: string;
    CallbackNumber: string;
    UserName: string;
    Password: string;
    Domain: string;
    lbHasPassword: Longbool;
    hRasConn: THRasConn;
    rasConnState: TRASConnState;
    rasConnError: Longint;
    DeviceType: string;
    DeviceName: string;
    rasMonitor: TRASMonitorThread;
    rasMonitorActive: Boolean;
    IsConnecting: Boolean;
  end;

type
  TRASConnectionMonitorThread = class(TThread)
  private
    { Private declarations }
    FOwner: TComponent;
    FEntryIndex: integer;
    FRASState: TRasConnState;
    FRASError: Longint;
    RefreshRate: Word;
  protected
    { Protected declarations }
    procedure Execute; override;
    procedure RASStatusChange;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent; RR: Word);
    destructor Destroy; override;
  end;

type
  TRASControl = class(TComponent)
  private
    { Private declarations }
    FOnStatus: TRASNotifyEvent;
    FOnEntryListChange: TNotifyEvent;
    FConnecting: Boolean;
    FRefreshRate: Word;
    FConnectionThread: TRASConnectionMonitorThread;
    FEntries: TStrings;
    procedure RaiseRASControlError(ErrorMessage: string);
    procedure ValidatEntryIndex(Index: integer);
    procedure SetMonitorConn(Value: Boolean);
    function GetIsConnecting(Index: integer): Boolean;
    function GetEntryName(Index: integer): string;
    function GetPhoneNumber(Index: integer): string;
    procedure SetPhoneNumber(Index: integer; Value: string);
    function GetCallbackNumber(Index: integer): string;
    procedure SetCallbackNumber(Index: integer; Value: string);
    function GetUserName(Index: integer): string;
    procedure SetUserName(Index: integer; Value: string);
    function GetPassword(Index: integer): string;
    procedure SetPassword(Index: integer; Value: string);
    function GetDomain(Index: integer): string;
    procedure SetDomain(Index: integer; Value: string);
    function GetDeviceType(Index: integer): string;
    function GetDeviceName(Index: integer): string;
    function GetHasPassword(Index: integer): Longbool;
    procedure SetHasPassword(Index: integer; RemovePW: Longbool);
    function GethRasConn(Index: integer): THRASConn;
    function GetrasConnState(Index: integer): TRASConnState;
    function GetrasConnError(Index: integer): Longint;
    function GetrasMonitor(Index: integer): TRASMonitorThread;
    function GetrasMonitorActive(Index: integer): Boolean;
    procedure SetRefreshRate(Value: Word);
  protected
    { Protected declarations }
    FMonitorConn: Boolean;
    procedure DoConnectionStatusChange(Sender: TObject; entryindex: integer; state: TRasConnState; error: Longint);
    procedure DoEntryListChange;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DialEntry(Index: integer);
    procedure HangupEntry(Index: integer);
    procedure SetDialParams(Index: integer; RemovePW: Boolean);
    procedure OSEditEntry(Index: integer);
    procedure OSAddEntry;
    procedure LoadRASEntryList;
    procedure RefreshRASEntry(Index: integer);
    function StatusString(state: TRasConnState): String;
    function ErrorString(error: Longint): String;
    property Entries: TStrings read FEntries;
    property Connecting[Index: integer]: Boolean read GetIsConnecting;
    property EntryName[Index: integer]: string read GetEntryName;
    property PhoneNumber[Index: integer]: string read GetPhoneNumber write SetPhoneNumber;
    property CallbackNumber[Index: integer]: string read GetCallbackNumber write SetCallbackNumber;
    property UserName[Index: integer]: string read GetUserName write SetUserName ;
    property Password[Index: integer]: string read GetPassword write SetPassword ;
    property Domain[Index: integer]: string read GetDomain write SetDomain ;
    property DeviceType[Index: integer]: string read GetDeviceType;
    property DeviceName[Index: integer]: string read GetDeviceName;
    property lbHasPassword[Index: integer]: Longbool read GetHasPassword write SetHasPassword;
    property hRasConn[Index: integer]: THRasConn read GethRasConn;
    property rasConnState[Index: integer]: TRASConnState read GetrasConnState;
    property rasConnError[Index: integer]: Longint read GetrasConnError;
    property rasMonitor[Index: integer]: TRASMonitorThread read GetrasMonitor;
    property rasMonitorActive[Index: integer]: Boolean read GetrasMonitorActive;
  published
    { Published declarations }
    property MonitorConnections: Boolean read FMonitorConn write SetMonitorConn;
    property RefreshRate: Word read FRefreshRate write SetRefreshRate;
    property Tag;
    property OnConnectionStatus: TRASNotifyEvent read FOnStatus write FOnStatus;
    property OnEntryListChange: TNotifyEvent read FOnEntryListChange write FOnEntryListChange;
  end;

procedure Register;

implementation

var
  FEntryList: TList;

procedure Register;
begin
  RegisterComponents('Custom', [TRASControl]);
end;

function IsEntryConnected(entryname: PChar; var rasConn: TRASConnStatus): THRASConn;
var
  bufsize: Longint;
  res, numEntries: Longint;
  x: Integer;
  stat: TRasConnStatus;
  entries: Array[1..100] of TRasConn;
  c: array [0..255] of Char;
begin
  Result := 0;
  ZeroMemory(@rasConn,SizeOf(rasConn));
  entries[1].dwSize := SizeOf(TRasConn);
  bufsize := SizeOf(TRasConn) * 100;
  FillChar(stat, Sizeof(TRasConnStatus), 0);
  stat.dwSize := Sizeof(TRasConnStatus);
  res := RasEnumConnections(@entries[1], bufsize, numEntries);
  if res = 0 then begin
    if numEntries > 0 then begin
      for x := 1 to numEntries do begin
        if StrComp(entries[x].szEntryName,entryname) = 0 then begin
          RASGetConnectStatus(entries[x].hrasconn,stat);
          Result := entries[x].hrasconn;
          CopyMemory(@rasConn,@entries[x],SizeOf(rasConn));
        end else begin
          rasConn.rasconnstate := RASCS_Disconnected;
          rasConn.dwError := 0;
        end;
      end;
    end else begin
      rasConn.rasconnstate := RASCS_Disconnected;
      rasConn.dwError := 0;
    end;
  end else begin
    ZeroMemory(@c,SizeOf(c));
    case res of
        ERROR_BUFFER_TOO_SMALL  : StrPCopy(c,'ERROR_BUFFER_TOO_SMALL');
        ERROR_NOT_ENOUGH_MEMORY : StrPCopy(c,'ERROR_NOT_ENOUGH_MEMORY');
      else
      RasGetErrorString(res, c, 256);
    end;
    raise ERASControlError.Create('RAS: ' + StrPas(c));
  end;
end;

procedure LoadEntryItem(Index: integer);
var
  res: integer;
  entry: PRASEntryItem;
  status: TRasConnStatus;
  dialparams: TRASDialParams;
  rasentry: LPRASEntry;
  entrysize: Longint;
  devsize: Longint;
  entries: array [0..99] of TRasConn;
begin
  entry := FEntryList.List[Index];
  // initialize rasentry
  rasentry := AllocMem(SizeOf(TRasEntry));
  ZeroMemory(rasentry,SizeOf(TRasEntry));
  rasentry^.dwSize := SizeOf(TRasEntry);
  entrysize := SizeOf(TRasEntry);
  devsize := 0;
  // initialize dialparams
  ZeroMemory(@dialparams,SizeOf(dialparams));
  dialparams.dwSize := SizeOf(dialparams);
  StrPCopy(dialparams.szEntryName,entry^.EntryName);
  // initialize status
  ZeroMemory(@status,SizeOf(status));
  status.dwSize := SizeOf(status);
  // initialize connection entries
  ZeroMemory(@entries[0],SizeOf(entries[0]));
  entries[0].dwSize := SizeOf(entries[0]);

  RasGetEntryDialParams(nil,dialparams,entry^.lbHasPassword);

  res := RasGetEntryProperties(nil,dialparams.szEntryName,
                               rasentry,entrysize,
                               nil,devsize);

  if res = ERROR_BUFFER_TOO_SMALL then begin
    ReallocMem(rasentry, entrysize);
    rasentry^.dwSize := SizeOf(TRASEntry);
    RasGetEntryProperties(nil,dialparams.szEntryName,
                          rasentry,entrysize,
                          nil,devsize);
  end;

  entry^.EntryName := StrPas(dialparams.szEntryName);
  entry^.PhoneNumber := StrPas(rasentry^.szLocalPhoneNumber);
  entry^.CallbackNumber := StrPas(dialparams.szCallbackNumber);
  entry^.UserName := StrPas(dialparams.szUserName);
  entry^.Password := StrPas(dialparams.szPassword);
  entry^.Domain := StrPas(dialparams.szDomain);
  entry^.DeviceType := StrPas(rasentry^.szDeviceType);
  entry^.DeviceName := StrPas(rasentry^.szDeviceName);

  FreeMem(rasentry);
end;

procedure RASCallBackProc1(hConn: THRasConn; msg: Integer; state: TRasConnState; dwError: Longint; dwEexterror: Longint); stdcall;
var
  x: integer;
  entry: PRASEntryItem;
begin
  for x := 0 to Pred(FEntryList.Count) do begin
    entry := FEntryList.List[x];
    LoadEntryItem(x);
    if entry^.hRasConn = hConn then begin
      entry^.rasConnState := state;
      entry^.rasConnError := dwError;
      entry^.rasMonitor.UpdateStatus;
    end;
  end;
end;

{ TRASMonitorThread }

procedure TRASMonitorThread.Execute;
var
  entry: PRASEntryItem;
begin
  while not Terminated do begin
    Sleep(50);
  end;
  entry := FEntryList.List[FEntryIndex];
  entry^.rasMonitorActive := False;
end;

procedure TRASMonitorThread.RASStatusChange;
begin
  (FOwner as TRASControl).DoConnectionStatusChange(Self,FEntryIndex,FRASState,FRASError);
end;

constructor TRASMonitorThread.Create(AOwner: TComponent; FSuspended: Boolean; Index: integer);
var
  entry: PRASEntryItem;
begin
  inherited Create(FSuspended);
  FreeOnTerminate := True;
  FOwner := AOwner;
  FEntryIndex := Index;
  entry := FEntryList.List[FEntryIndex];
  RASCallbackProc := @RASCallBackProc1;
  entry^.rasMonitorActive := True;
end;

destructor TRASMonitorThread.Destroy;
begin
  inherited Destroy;
end;

procedure TRASMonitorThread.UpdateStatus;
var
  entry: PRASEntryItem;
begin
  entry := FEntryList.List[FEntryIndex];
  FRASState := entry^.rasConnState;
  FRASError := entry^.rasConnError;
  Synchronize(RASStatusChange);
end;

{ TRASConnectionMonitorThread }

constructor TRASConnectionMonitorThread.Create(AOwner: TComponent; RR: Word);
begin
  inherited Create(False);
  FreeOnTerminate := True;
  FOwner := AOwner;
  RefreshRate := RR;
end;

destructor TRASConnectionMonitorThread.Destroy;
begin
  inherited Destroy;
end;

procedure TRASConnectionMonitorThread.Execute;
var
  x : Integer;
  entry: PRASEntryItem;
  LastState: TRASConnState;
  status: TRASConnStatus;
  LastError: Longint;
begin
  while not Terminated do begin
    for x := 0 to (FEntryList.Count - 1) do begin
      entry := FEntryList.List[x];
      if not entry^.rasMonitorActive then begin
        LastState := entry^.rasConnState;
        LastError := entry^.rasConnError;
        LoadEntryItem(x);
        ZeroMemory(@status,SizeOf(status));
        status.dwSize := SizeOf(status);
        entry^.hRasConn := IsEntryConnected(PChar(entry^.EntryName),status);
        entry^.rasConnState := status.rasconnstate;
        entry^.rasConnError := status.dwError;
        if (LastState <> entry^.rasConnState) or (LastError <> entry^.rasConnError) then begin
          FEntryIndex := x;
          FRASState := entry^.rasConnState;
          FRASError := entry^.rasConnError;
          Synchronize(RASStatusChange);
        end;
      end;
    end;
    Sleep(RefreshRate + 10);
  end;

  for x := 0 to Pred(FEntryList.Count) do begin
    entry := FEntryList.List[x];
    if not entry^.rasMonitorActive then begin
      entry^.rasConnState := RASControl_Unknown;
      entry^.rasConnError := 0;
      entry^.IsConnecting := False;
      entry^.hRasConn := 0;
      FEntryIndex := x;
      FRASState := entry^.rasConnState;
      FRASError := entry^.rasConnError;
      Synchronize(RASStatusChange);
    end;
  end;

  (FOwner as TRASControl).FMonitorConn := False;
end;

procedure TRASConnectionMonitorThread.RASStatusChange;
begin
  (FOwner as TRASControl).DoConnectionStatusChange(Self,FEntryIndex,FRASState,FRASError);
end;

{ TRASControl }

    { PRIVATE }

procedure TRASControl.RaiseRASControlError(ErrorMessage: string);
begin
  raise ERASControlError.Create('('+Self.ClassName+') ' + ErrorMessage);
end;

procedure TRASControl.ValidatEntryIndex(Index: integer);
begin
  if (Index < 0) or (Index > (FEntryList.Count - 1)) then
    RaiseRASControlError('Index ('+IntToStr(Index)+') out of bounds.');
end;

procedure TRASControl.SetMonitorConn(Value: Boolean);
begin
  if not (csDesigning in ComponentState) then begin
    if Value <> FMonitorConn then begin
      if Value then begin
        FConnectionThread := TRASConnectionMonitorThread.Create(Self, FRefreshRate);
      end else begin
        if Assigned(FConnectionThread) then
          FConnectionThread.Terminate;
      end;
    end;
  end;
  FMonitorConn := Value;
end;

function TRASControl.GetIsConnecting(Index: integer): Boolean;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.IsConnecting;
end;

function TRASControl.GetEntryName(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.EntryName;
end;

function TRASControl.GetPhoneNumber(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.PhoneNumber;
end;

procedure TRASControl.SetPhoneNumber(Index: integer; Value: string);
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  if Length(Value) > RAS_MaxPhoneNumber then begin
    RaiseRASControlError('Value assigned to PhoneNumber is longer than '+IntToStr(RAS_MaxPhoneNumber));
  end;
  entry := FEntryList.List[Index];
  if Value <> entry^.PhoneNumber then
    entry^.PhoneNumber := Value;
end;

function TRASControl.GetCallbackNumber(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.CallbackNumber;
end;

procedure TRASControl.SetCallbackNumber(Index: integer; Value: string);
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  if Length(Value) > RAS_MaxCallbackNumber then begin
    RaiseRASControlError('Value assigned to CallbackNumber is longer than '+IntToStr(RAS_MaxCallbackNumber));
  end;
  entry := FEntryList.List[Index];
  if Value <> entry^.CallbackNumber then
    entry^.CallbackNumber := Value;
end;

function TRASControl.GetUserName(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.UserName;
end;

procedure TRASControl.SetUserName(Index: integer; Value: string);
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  if Length(Value) > UNLEN then begin
    RaiseRASControlError('Value assigned to UserName is longer than '+IntToStr(UNLEN));
  end;
  entry := FEntryList.List[Index];
  if Value <> entry^.UserName then
    entry^.UserName := Value;
end;

function TRASControl.GetPassword(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.Password;
end;

procedure TRASControl.SetPassword(Index: integer; Value: string);
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  if Length(Value) > PWLEN then begin
    RaiseRASControlError('Value assigned to Password is longer than '+IntToStr(PWLEN));
  end;
  entry := FEntryList.List[Index];
  if Value <> entry^.Password then
    entry^.Password := Value;
end;

function TRASControl.GetDomain(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.Domain;
end;

procedure TRASControl.SetDomain(Index: integer; Value: string);
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  if Length(Value) > DNLEN then begin
    RaiseRASControlError('Value assigned to Domain is longer than '+IntToStr(DNLEN));
  end;
  entry := FEntryList.List[Index];
  if Value <> entry^.Domain then
    entry^.Domain := Value;
end;

function TRASControl.GetDeviceType(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.DeviceType;
end;

function TRASControl.GetDeviceName(Index: integer): string;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.DeviceName;
end;

function TRASControl.GetHasPassword(Index: integer): Longbool;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.lbHasPassword;
end;

procedure TRASControl.SetHasPassword(Index: integer; RemovePW: Longbool);
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  entry^.lbHasPassword := RemovePW;
end;

function TRASControl.GethRasConn(Index: integer): THRASConn;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.hRasConn;
end;

function TRASControl.GetrasConnState(Index: integer): TRASConnState;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.rasConnState;
end;

function TRASControl.GetrasConnError(Index: integer): Longint;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.rasConnError;
end;

function TRASControl.GetrasMonitor(Index: integer): TRASMonitorThread;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.rasMonitor;
end;

function TRASControl.GetrasMonitorActive(Index: integer): Boolean;
var
  entry: PRASEntryItem;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  Result := entry^.rasMonitorActive;
end;

procedure TRASControl.SetRefreshRate(Value: Word);
begin
  if Value <> FRefreshRate then begin
    FRefreshRate := Value;
    if not (csDesigning in ComponentState) then
      if FMonitorConn  then
        FConnectionThread.RefreshRate := Value;
  end;
end;

    { PROTECTED }

procedure TRASControl.DoConnectionStatusChange(Sender: TObject; entryindex: integer; state: TRasConnState; error: Longint);
var
  entry: PRASEntryItem;
begin
  entry := FEntryList.List[entryindex];
  entry^.rasConnState := state;
  case state of
    RASCS_Connected    : entry^.IsConnecting := False;
    RASCS_Disconnected : entry^.IsConnecting := False;
    RASControl_Unknown : entry^.IsConnecting := False;
  else
    entry^.IsConnecting := True;
  end;
  FConnecting := entry^.IsConnecting;
  if Assigned(FOnStatus) then begin
    FOnStatus( Sender, entryindex, state, error);
  end;
end;

procedure TRASControl.DoEntryListChange;
begin
  if Assigned(FOnEntryListChange) then
    FOnEntryListChange(Self);
end;

    { PUBLIC }

constructor TRASControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEntryList := TList.Create;
  FEntryList.Capacity := 10;
  FRefreshRate := 1000;
  FEntries := TStringList.Create;
  LoadRASEntryList;
  SetMonitorConn(FMonitorConn);
end;

destructor TRASControl.Destroy;
var
  entry: PRASEntryItem;
  x: integer;
begin
  FEntries.Free;
  if Assigned(FEntryList) then begin
    for x := 0 to (FEntryList.Count - 1) do begin
      entry := FEntryList.List[x];
      if entry^.rasMonitorActive then entry^.rasMonitor.Terminate;
      Dispose(entry);
    end;
    FEntryList.Free;
    FEntryList := nil;
  end;
  inherited Destroy;
end;

procedure TRASControl.DialEntry(Index: integer);
var
  c: array [0..255] of Char;
  res: DWORD;
  entry: PRASEntryItem;
  dialparams: TRASDialParams;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  if FConnecting then begin
    RaiseRASControlError('A connection attempt is already in progress.');
  end;
  if (entry^.rasConnState = RASCS_Disconnected) or
     (entry^.rasConnState = RASControl_Unknown) then begin
    entry^.IsConnecting := True;
    FConnecting := True;
    ZeroMemory(@dialparams,SizeOf(dialparams));
    dialparams.dwSize := SizeOf(dialparams);
    StrPCopy(dialparams.szEntryName      ,entry^.EntryName);
    StrPCopy(dialparams.szPhoneNumber    ,entry^.PhoneNumber);
    StrPCopy(dialparams.szCallbackNumber ,entry^.CallbackNumber);
    StrPCopy(dialparams.szUserName       ,entry^.UserName);
    StrPCopy(dialparams.szPassword       ,entry^.Password);
    StrPCopy(dialparams.szDomain         ,entry^.Domain);

    res := RASSetEntryDialParams(nil,dialparams,False);
    if res <> 0 then begin
      RasGetErrorString(res, c, 256);
      HangUpEntry(Index);
      RaiseRASControlError('Dial failed with message: '+StrPas(c));
    end;

    entry^.RASMonitor := TRASMonitorThread.Create(Self,False,Index);
    res := RasDial(nil,nil,
                   dialparams,1,
                   entry^.rasMonitor.RASCallBackProc,entry^.hRasConn);

    if res <> 0 then begin
      entry^.RASMonitor.Terminate;
      HangUpEntry(Index);
      ZeroMemory(@c,SizeOf(c));
      case res of
           ERROR_NOT_ENOUGH_MEMORY : StrPCopy(c,'Not enough memory.');
        else
        RasGetErrorString(res, c, 256);
      end;
      RaiseRASControlError('Dial failed. '+Trim(StrPas(c)));
    end;
  end else begin
    RaiseRASControlError('Entry is already connected.');
  end;
end;

procedure TRASControl.HangupEntry(Index: integer);
var
  entry: PRASEntryItem;
  c: array [0..255] of Char;
  res: DWORD;
begin
  ValidatEntryIndex(Index);
  entry := FEntryList.List[Index];
  if entry^.rasConnState <> RASCS_Disconnected then begin
    if entry^.rasMonitorActive then
      entry^.rasMonitor.Terminate;
    res := RasHangUp(entry^.hRasConn);
    if res <> 0 then begin
      ZeroMemory(@c,SizeOf(c));
      case res  of
        ERROR_INVALID_HANDLE : StrPCopy(c,'Invalid handle.');
        else
          RasGetErrorString(res, c, 256);
      end;
      RaiseRASControlError('Hang up failed. '+Trim(StrPas(c)));
    end;
    entry^.hRasConn := 0;
    entry^.IsConnecting := False;
    entry^.rasConnState := RASCS_Disconnected;
    entry^.rasMonitor := nil;
    DoConnectionStatusChange(Self,Index,entry^.rasConnState,0);
  end;
end;

procedure TRASControl.SetDialParams(Index: integer; RemovePW: Boolean);
var
  c: array [0..255] of Char;
  res: DWORD;
  entry: PRASEntryItem;
  dialparams: TRASDialParams;
begin
  if (Index < 0) or (Index > Pred(FEntryList.Count)) then begin
    RaiseRASControlError('Index '+IntToStr(Index)+' out of bounds.');
  end;
  entry := FEntryList.List[Index];
  ZeroMemory(@dialparams,SizeOf(dialparams));
  dialparams.dwSize := SizeOf(dialparams);
  StrPCopy(dialparams.szEntryName      ,entry^.EntryName);
  StrPCopy(dialparams.szPhoneNumber    ,entry^.PhoneNumber);
  StrPCopy(dialparams.szCallbackNumber ,entry^.CallbackNumber);
  StrPCopy(dialparams.szUserName       ,entry^.UserName);
  StrPCopy(dialparams.szPassword       ,entry^.Password);
  StrPCopy(dialparams.szDomain         ,entry^.Domain);
  res := RASSetEntryDialParams(nil,dialparams,RemovePW);
  if res <> 0 then begin
    RasGetErrorString(res, c, 256);
    RaiseRASControlError('SetDialParams failed: '+StrPas(c));
  end;
end;

procedure TRASControl.OSEditEntry(Index: integer);
var
  entry: PRASEntryItem;
  c: array [0..255] of Char;
  res: DWORD;
begin
  if (Index < 0) or (Index > Pred(FEntryList.Count)) then begin
    RaiseRASControlError('Index '+IntToStr(Index)+' out of bounds.');
  end;
  entry := FEntryList.List[Index];
  res := RASEditPhonebookEntry(0,nil,PChar(entry^.EntryName));
  if res <> 0 then begin
    RasGetErrorString(res, c, 256);
    RaiseRASControlError('OSEditEntry failed. '+StrPas(c));
  end;
  LoadRASEntryList;
end;

procedure TRASControl.OSAddEntry;
var
  c: array [0..255] of Char;
  res: DWORD;
begin
  res := RASCreatePhonebookEntry(0,nil);
  if res <> 0 then begin
    RasGetErrorString(res, c, 256);
    RaiseRASControlError('OSAddEntry failed. '+StrPas(c));
  end;
  LoadRASEntryList;
end;

procedure TRASControl.LoadRASEntryList;
var
  c: array [0..255] of char;
  bufsize: Longint;
  numEntries: Longint;
  entries, p: LPRasEntryName;
  dialparams: TRASDialParams;
  connstatus: TRASConnStatus;
  x: Integer;
  res: Integer;
  NewItem: PRASEntryItem;
  rasentry: LPRASEntry;
  entrysize: Longint;
  devsize: Longint;
begin
  FEntryList.Clear;
  FEntries.Clear;
  entries := AllocMem(SizeOf(TRasEntryName));
  entries^.dwSize := SizeOf(TRasEntryName);
  bufsize := SizeOf(TRasEntryName);
  res := RasEnumEntries(nil, nil, entries, bufsize, numEntries);
  if res = ERROR_BUFFER_TOO_SMALL then begin
    ReallocMem(entries, bufsize);
    FillChar(entries^, bufsize, 0);
    entries^.dwSize := SizeOf(TRasEntryName);
    res := RasEnumEntries(nil, nil, entries, bufsize, numEntries);
  end;
  if res = 0 then begin
    if numEntries > 0 then begin
      p := entries;
      for x := 0 to (numEntries - 1) do begin
        // initialize RASEntryItem
        New(NewItem);
        ZeroMemory(@NewItem^,SizeOf(NewItem^));
        // initialize dialparams
        ZeroMemory(@dialparams,SizeOf(dialparams));
        dialparams.dwSize := SizeOf(dialparams);
        StrCopy(dialparams.szEntryName,p^.szEntryName);
        RasGetEntryDialParams(nil,dialparams,NewItem^.lbHasPassword);
        // initialize rasentry
        rasentry := AllocMem(SizeOf(TRasEntry));
        rasentry^.dwSize := SizeOf(TRASEntry);
        entrysize := SizeOf(TRasEntry);
        devsize := 0;
        res := RasGetEntryProperties(nil,dialparams.szEntryName,rasentry,entrysize,nil,devsize);
        if res = ERROR_BUFFER_TOO_SMALL then begin
          ReallocMem(rasentry, entrysize);
          rasentry^.dwSize := SizeOf(TRASEntry);
          RasGetEntryProperties(nil,dialparams.szEntryName,rasentry,entrysize,nil,devsize);
        end;
        // populate RASEntryItem
        NewItem^.EntryName := StrPas(dialparams.szEntryName);
        NewItem^.PhoneNumber := StrPas(rasentry.szLocalPhoneNumber);
        NewItem^.CallbackNumber := StrPas(dialparams.szCallbackNumber);
        NewItem^.UserName := StrPas(dialparams.szUserName);
        NewItem^.Password := StrPas(dialparams.szPassword);
        NewItem^.Domain := StrPas(dialparams.szDomain);
        if FMonitorConn then begin
          ZeroMemory(@connstatus,SizeOf(connstatus));
          connstatus.dwSize := SizeOf(connstatus);
          NewItem^.hRasConn := IsEntryConnected(dialparams.szEntryName,connstatus);
          NewItem^.rasConnError := connstatus.dwError;
          NewItem^.rasConnState := connstatus.rasconnstate;
          NewItem^.DeviceType := StrPas(connstatus.szDeviceType);
          NewItem^.DeviceName := StrPas(connstatus.szDeviceName);
        end else begin
          NewItem^.hRasConn := 0;
          NewItem^.rasConnState := RASControl_Unknown;
          NewItem^.rasConnError := 0;
          NewItem^.DeviceType := StrPas(rasentry^.szDeviceType);
          NewItem^.DeviceName := StrPas(rasentry^.szDeviceName);
        end;
        // add RASEntryItem to FEntryList
        FEntries.Add(p^.szEntryName);
        FEntryList.Add(NewItem);
        Inc(p);
      end;
    end;
  end else begin
    RasGetErrorString(res, c, 256);
  end;
  FreeMem(entries);
  FreeMem(rasentry);
  DoEntryListChange;
end;

procedure TRASControl.RefreshRASEntry(Index: integer);
var
  entry: PRASEntryItem;
begin
  if (Index < 0) or (Index > Pred(FEntryList.Count)) then begin
    RaiseRASControlError('Index '+IntToStr(Index)+' out of bounds.');
  end;
  entry := FEntryList.List[Index];
  LoadEntryItem(Index);
  DoConnectionStatusChange(Self, Index, entry^.rasConnState, entry^.rasConnError);
end;

function TRASControl.StatusString(state: TRasConnState): String;
var
  s: string;
begin
  s := '';
  case State of
      RASCS_OpenPort               : s := 'Opening port';
      RASCS_PortOpened             : s := 'Port opened';
      RASCS_ConnectDevice          : s := 'Connecting device';
      RASCS_DeviceConnected        : s := 'Device connected';
      RASCS_AllDevicesConnected    : s := 'All devices connected';
      RASCS_Authenticate           : s := 'Start authenticating';
      RASCS_AuthNotify             : s := 'Authentication: notify';
      RASCS_AuthRetry              : s := 'Authentication: retry';
      RASCS_AuthCallback           : s := 'Authentication: callback';
      RASCS_AuthChangePassword     : s := 'Authentication: change password';
      RASCS_AuthProject            : s := 'Authentication: projecting';
      RASCS_AuthLinkSpeed          : s := 'Authentication: link speed';
      RASCS_AuthAck                : s := 'Authentication: acknowledge';
      RASCS_ReAuthenticate         : s := 'Authentication: reauthenticate';
      RASCS_Authenticated          : s := 'Authenticated';
      RASCS_PrepareForCallback     : s := 'Preparing for callback';
      RASCS_WaitForModemReset      : s := 'Waiting for modem reset';
      RASCS_WaitForCallback        : s := 'Waiting for callback';
      RASCS_Projected              : s := 'Projected';
      RASCS_StartAuthentication    : s := 'Start authentication';
      RASCS_CallbackComplete       : s := 'Callback complete';
      RASCS_LogonNetwork           : s := 'Logging on network';
      RASCS_Interactive            : s := 'Interactive';
      RASCS_RetryAuthentication    : s := 'Retry Authentication';
      RASCS_CallbackSetByCaller    : s := 'Callback set by caller';
      RASCS_PasswordExpired        : s := 'Password expired';
      RASCS_Connected              : s := 'Connected';
      RASCS_Disconnected           : s := 'Disconnected';

      RASControl_Unknown           : s := 'Unknown';
  end;
  Result := s;
end;

function TRASControl.ErrorString(error: Longint): String;
var
  c: array [0..255] of Char;
begin
  ZeroMemory(@c,SizeOf(c));
  RasGetErrorString(error, c, 256);
  Result := c;
end;

end.
